@@ -2,8 +2,6 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file.
 
-// @dart = 2.6
-
 part of dart.collection;
 
 /// A [Queue] is a collection that can be manipulated at both ends. One
@@ -11,20 +9,37 @@ part of dart.collection;
 /// an [Iterator].
 ///
 /// It is generally not allowed to modify the queue (add or remove entries)
-/// while an operation on the queue is being performed, for example during a
+/// while an operation in the queue is being performed, for example during a
 /// call to [forEach].
 /// Modifying the queue while it is being iterated will most likely break the
 /// iteration.
 /// This goes both for using the [iterator] directly, or for iterating an
 /// `Iterable` returned by a method like [map] or [where].
-abstract class Queue<E> implements EfficientLengthIterable<E> {
+///
+/// Example:
+/// ```dart
+/// final queue = Queue<int>(); // ListQueue() by default
+/// print(queue.runtimeType); // ListQueue
+///
+/// // Adding items to queue
+/// queue.addAll([1, 2, 3]);
+/// queue.addFirst(0);
+/// queue.addLast(10);
+/// print(queue); // {0, 1, 2, 3, 10}
+///
+/// // Removing items from queue
+/// queue.removeFirst();
+/// queue.removeLast();
+/// print(queue); // {1, 2, 3}
+/// ```
+abstract interface class Queue<E> implements EfficientLengthIterable<E> {
   /// Creates a queue.
   factory Queue() = ListQueue<E>;
 
   /// Creates a queue containing all [elements].
   ///
   /// The element order in the queue is as if the elements were added using
-  /// [addLast] in the order provided by [elements.iterator].
+  /// [addLast] in the order provided by [elements].iterator.
   ///
   /// All the [elements] should be instances of [E].
   /// The `elements` iterable itself may have any element type, so this
@@ -32,14 +47,14 @@ abstract class Queue<E> implements EfficientLengthIterable<E> {
   /// ```dart
   /// Queue<SuperType> superQueue = ...;
   /// Queue<SubType> subQueue =
-  ///     new Queue<SubType>.from(superQueue.whereType<SubType>());
+  ///     Queue<SubType>.from(superQueue.whereType<SubType>());
   /// ```
   factory Queue.from(Iterable elements) = ListQueue<E>.from;
 
   /// Creates a queue from [elements].
   ///
   /// The element order in the queue is as if the elements were added using
-  /// [addLast] in the order provided by [elements.iterator].
+  /// [addLast] in the order provided by [elements].iterator.
   factory Queue.of(Iterable<E> elements) = ListQueue<E>.of;
 
   /// Adapts [source] to be a `Queue<T>`.
@@ -47,13 +62,17 @@ abstract class Queue<E> implements EfficientLengthIterable<E> {
   /// Any time the queue would produce an element that is not a [T],
   /// the element access will throw.
   ///
-  /// Any time a [T] value is attempted stored into the adapted queue,
-  /// the store will throw unless the value is also an instance of [S].
+  /// When a [T] value is stored into the adapted queue,
+  /// the operation will throw unless the value is also an instance of [S].
   ///
   /// If all accessed elements of [source] are actually instances of [T],
-  /// and if all elements stored into the returned queue are actually instance
+  /// and if all elements stored into the returned queue are actually instances
   /// of [S],
   /// then the returned queue can be used as a `Queue<T>`.
+  ///
+  /// Methods which accept `Object?` as argument, like [contains] and [remove],
+  /// will pass the argument directly to this queue's method
+  /// without any checks.
   static Queue<T> castFrom<S, T>(Queue<S> source) => CastQueue<S, T>(source);
 
   /// Provides a view of this queue as a queue of [R] instances, if necessary.
@@ -63,9 +82,15 @@ abstract class Queue<E> implements EfficientLengthIterable<E> {
   /// that is not an instance of [R], the access will throw instead.
   ///
   /// Elements added to the queue (e.g., by using [addFirst] or [addAll])
-  /// must be instance of [R] to be valid arguments to the adding function,
-  /// and they must be instances of [E] as well to be accepted by
+  /// must be instances of [R] to be valid arguments to the adding function,
+  /// and they must also be instances of [E] to be accepted by
   /// this queue as well.
+  ///
+  /// Methods which accept `Object?` as argument, like [contains] and [remove],
+  /// will pass the argument directly to the this queue's method
+  /// without any checks.
+  /// That means that you can do `queueOfStrings.cast<int>().remove("a")`
+  /// successfully, even if it looks like it shouldn't have any effect.
   Queue<R> cast<R>();
 
   /// Removes and returns the first element of this queue.
@@ -87,11 +112,11 @@ abstract class Queue<E> implements EfficientLengthIterable<E> {
   /// Adds [value] at the end of the queue.
   void add(E value);
 
-  /// Remove a single instance of [value] from the queue.
+  /// Removes a single instance of [value] from the queue.
   ///
   /// Returns `true` if a value was removed, or `false` if the queue
   /// contained no element equal to [value].
-  bool remove(Object value);
+  bool remove(Object? value);
 
   /// Adds all elements of [iterable] at the end of the queue. The
   /// length of the queue is extended by the length of [iterable].
@@ -111,106 +136,71 @@ abstract class Queue<E> implements EfficientLengthIterable<E> {
   void clear();
 }
 
-class _DoubleLink<Link extends _DoubleLink<Link>> {
-  Link _previousLink;
-  Link _nextLink;
-
-  void _link(Link previous, Link next) {
-    _nextLink = next;
-    _previousLink = previous;
-    if (previous != null) previous._nextLink = this;
-    if (next != null) next._previousLink = this;
-  }
-
-  void _unlink() {
-    if (_previousLink != null) _previousLink._nextLink = _nextLink;
-    if (_nextLink != null) _nextLink._previousLink = _previousLink;
-    _nextLink = null;
-    _previousLink = null;
-  }
-}
-
-/// An entry in a doubly linked list. It contains a pointer to the next
-/// entry, the previous entry, and the boxed element.
-class DoubleLinkedQueueEntry<E> extends _DoubleLink<DoubleLinkedQueueEntry<E>> {
-  /// The element in the queue.
-  E element;
-
-  DoubleLinkedQueueEntry(this.element);
-
-  /// Appends the given [e] as entry just after this entry.
-  void append(E e) {
-    DoubleLinkedQueueEntry<E>(e)._link(this, _nextLink);
-  }
-
-  /// Prepends the given [e] as entry just before this entry.
-  void prepend(E e) {
-    DoubleLinkedQueueEntry<E>(e)._link(_previousLink, this);
-  }
-
-  E remove() {
-    _unlink();
-    return element;
-  }
-
-  /// Returns the previous entry or `null` if there is none.
-  DoubleLinkedQueueEntry<E> previousEntry() => _previousLink;
-
-  /// Returns the next entry or `null` if there is none.
-  DoubleLinkedQueueEntry<E> nextEntry() => _nextLink;
-}
-
-/// Interface for the link classes used by [DoubleLinkedQueue].
+/// Interface and base class for the link classes used by [DoubleLinkedQueue].
 ///
 /// Both the [_DoubleLinkedQueueElement] and [_DoubleLinkedQueueSentinel]
 /// implement this interface.
-/// The entry contains a link back to the queue, so calling `append`
-/// or `prepend` can correctly update the element count.
-abstract class _DoubleLinkedQueueEntry<E> extends DoubleLinkedQueueEntry<E> {
-  DoubleLinkedQueue<E> _queue;
-  _DoubleLinkedQueueEntry(E element, this._queue) : super(element);
+abstract class _DoubleLinkedQueueEntry<E> {
+  _DoubleLinkedQueueEntry<E>? _previousLink;
+  _DoubleLinkedQueueEntry<E>? _nextLink;
 
-  DoubleLinkedQueueEntry<E> _asNonSentinelEntry();
-
-  void _append(E e) {
-    _DoubleLinkedQueueElement<E>(e, _queue)._link(this, _nextLink);
+  void _link(
+      _DoubleLinkedQueueEntry<E>? previous, _DoubleLinkedQueueEntry<E>? next) {
+    _nextLink = next;
+    _previousLink = previous;
+    previous?._nextLink = this;
+    next?._previousLink = this;
   }
 
-  void _prepend(E e) {
-    _DoubleLinkedQueueElement<E>(e, _queue)._link(_previousLink, this);
+  void _unlink() {
+    _previousLink?._nextLink = _nextLink;
+    _nextLink?._previousLink = _previousLink;
+    _previousLink = _nextLink = null;
+  }
+
+  _DoubleLinkedQueueElement<E>? _asNonSentinelEntry();
+
+  void _append(E element, DoubleLinkedQueue<E>? queue) {
+    _DoubleLinkedQueueElement<E>(element, queue)._link(this, _nextLink);
+  }
+
+  void _prepend(E element, DoubleLinkedQueue<E>? queue) {
+    _DoubleLinkedQueueElement<E>(element, queue)._link(_previousLink, this);
   }
 
   E _remove();
 
-  E get _element => element;
-
-  DoubleLinkedQueueEntry<E> nextEntry() {
-    _DoubleLinkedQueueEntry<E> entry = _nextLink;
-    return entry._asNonSentinelEntry();
-  }
-
-  DoubleLinkedQueueEntry<E> previousEntry() {
-    _DoubleLinkedQueueEntry<E> entry = _previousLink;
-    return entry._asNonSentinelEntry();
-  }
+  E get element;
 }
 
-/// The actual entry type used by the [DoubleLinkedQueue].
+/// Linked list entry used by the [DoubleLinkedQueue] to hold an element.
 ///
-/// The entry contains a reference to the queue, allowing
+/// These entry objects are also exposed by [DoubleLinkedQueue.firstEntry],
+/// [DoubleLinkedQueue.lastEntry] and [DoubleLinkedQueue.forEachEntry].
+///
+/// The entry contains both the [element] (which is mutable to anyone with
+/// access to the entry object) and a reference to the queue, allowing
 /// [append]/[prepend] to update the list length.
-class _DoubleLinkedQueueElement<E> extends _DoubleLinkedQueueEntry<E> {
-  _DoubleLinkedQueueElement(E element, DoubleLinkedQueue<E> queue)
-      : super(element, queue);
+///
+/// When an entry is removed from its queue, the [_queue] is set to `null`
+/// and will never change again. You can still use the unlinked entry
+/// to create a new list, by calling [append] and [prepend], but it won't
+/// be part of any [DoubleLinkedQueue].
+class _DoubleLinkedQueueElement<E> extends _DoubleLinkedQueueEntry<E>
+    implements DoubleLinkedQueueEntry<E> {
+  DoubleLinkedQueue<E>? _queue;
+  E element;
+
+  _DoubleLinkedQueueElement(this.element, this._queue);
 
   void append(E e) {
-    _append(e);
-    if (_queue != null) _queue._elementCount++;
+    _append(e, _queue);
+    _queue?._elementCount++;
   }
 
   void prepend(E e) {
-    _prepend(e);
-    if (_queue != null) _queue._elementCount++;
+    _prepend(e, _queue);
+    _queue?._elementCount++;
   }
 
   E _remove() {
@@ -220,30 +210,43 @@ class _DoubleLinkedQueueElement<E> extends _DoubleLinkedQueueEntry<E> {
   }
 
   E remove() {
-    if (_queue != null) _queue._elementCount--;
+    _queue?._elementCount--;
     return _remove();
   }
 
-  _DoubleLinkedQueueElement<E> _asNonSentinelEntry() {
-    return this;
-  }
+  _DoubleLinkedQueueElement<E> _asNonSentinelEntry() => this;
+
+  DoubleLinkedQueueEntry<E>? previousEntry() =>
+      _previousLink?._asNonSentinelEntry();
+
+  DoubleLinkedQueueEntry<E>? nextEntry() => _nextLink?._asNonSentinelEntry();
 }
 
-/// A sentinel in a double linked list is used to manipulate the list
-/// at both ends.
-/// A double linked list has exactly one sentinel,
+/// A header object used to hold the two ends of a double linked queue.
+///
+/// A [DoubleLinkedQueue] has exactly one sentinel,
 /// which is the only entry when the list is constructed.
-/// Initially, a sentinel has its next and previous entry point to itself.
-/// A sentinel does not box any user element.
+///
+/// Initially, a sentinel has its next and previous entries point to itself.
+/// Its next and previous links are never `null` after creation, and
+/// the entries linked always form a circular structure with the next link
+/// pointing to the first element of the queue, and the previous link
+/// pointing to the last element of the queue, or both pointing to itself
+/// again if the queue becomes empty.
+///
+/// Implements [_DoubleLinkedQueueEntry._remove] and
+/// [_DoubleLinkedQueueEntry.element] as throwing because
+/// it makes it simple to implement members like [Queue.removeFirst]
+/// or [Queue.first] as throwing on an empty queue.
+///
+/// A sentinel does not contain any user element.
 class _DoubleLinkedQueueSentinel<E> extends _DoubleLinkedQueueEntry<E> {
-  _DoubleLinkedQueueSentinel(DoubleLinkedQueue<E> queue) : super(null, queue) {
+  _DoubleLinkedQueueSentinel() {
     _previousLink = this;
     _nextLink = this;
   }
 
-  DoubleLinkedQueueEntry<E> _asNonSentinelEntry() {
-    return null;
-  }
+  Null _asNonSentinelEntry() => null;
 
   /// Hit by, e.g., [DoubleLinkedQueue.removeFirst] if the queue is empty.
   E _remove() {
@@ -251,7 +254,7 @@ class _DoubleLinkedQueueSentinel<E> extends _DoubleLinkedQueueEntry<E> {
   }
 
   /// Hit by, e.g., [DoubleLinkedQueue.first] if the queue is empty.
-  E get _element {
+  E get element {
     throw IterableElementError.noElement();
   }
 }
@@ -259,31 +262,31 @@ class _DoubleLinkedQueueSentinel<E> extends _DoubleLinkedQueueEntry<E> {
 /// A [Queue] implementation based on a double-linked list.
 ///
 /// Allows constant time add, remove-at-ends and peek operations.
-class DoubleLinkedQueue<E> extends Iterable<E> implements Queue<E> {
-  _DoubleLinkedQueueSentinel<E> _sentinel;
+final class DoubleLinkedQueue<E> extends Iterable<E> implements Queue<E> {
+  final _DoubleLinkedQueueSentinel<E> _sentinel =
+      _DoubleLinkedQueueSentinel<E>();
+
   int _elementCount = 0;
 
-  DoubleLinkedQueue() {
-    _sentinel = _DoubleLinkedQueueSentinel<E>(this);
-  }
+  DoubleLinkedQueue();
 
   /// Creates a double-linked queue containing all [elements].
   ///
   /// The element order in the queue is as if the elements were added using
-  /// [addLast] in the order provided by [elements.iterator].
+  /// [addLast] in the order provided by [elements].iterator.
   ///
   /// All the [elements] should be instances of [E].
-  /// The `elements` iterable itself may have any element type, so this
-  /// constructor can be used to down-cast a `Queue`, for example as:
+  /// The [elements] iterable itself may have any element type, so this
+  /// constructor can be used to down-cast a [Queue], for example as:
   /// ```dart
   /// Queue<SuperType> superQueue = ...;
   /// Queue<SubType> subQueue =
-  ///     new DoubleLinkedQueue<SubType>.from(superQueue.whereType<SubType>());
+  ///     DoubleLinkedQueue<SubType>.from(superQueue.whereType<SubType>());
   /// ```
-  factory DoubleLinkedQueue.from(Iterable elements) {
-    Queue<E> list = DoubleLinkedQueue<E>();
+  factory DoubleLinkedQueue.from(Iterable<dynamic> elements) {
+    DoubleLinkedQueue<E> list = DoubleLinkedQueue<E>();
     for (final e in elements) {
-      list.addLast(e);
+      list.addLast(e as E);
     }
     return list;
   }
@@ -291,54 +294,57 @@ class DoubleLinkedQueue<E> extends Iterable<E> implements Queue<E> {
   /// Creates a double-linked queue from [elements].
   ///
   /// The element order in the queue is as if the elements were added using
-  /// [addLast] in the order provided by [elements.iterator].
+  /// [addLast] in the order provided by [elements].iterator.
   factory DoubleLinkedQueue.of(Iterable<E> elements) =>
       DoubleLinkedQueue<E>()..addAll(elements);
 
   Queue<R> cast<R>() => Queue.castFrom<E, R>(this);
+
   int get length => _elementCount;
 
   void addLast(E value) {
-    _sentinel._prepend(value);
+    _sentinel._prepend(value, this);
     _elementCount++;
   }
 
   void addFirst(E value) {
-    _sentinel._append(value);
+    _sentinel._append(value, this);
     _elementCount++;
   }
 
   void add(E value) {
-    _sentinel._prepend(value);
+    _sentinel._prepend(value, this);
     _elementCount++;
   }
 
   void addAll(Iterable<E> iterable) {
     for (final E value in iterable) {
-      _sentinel._prepend(value);
+      _sentinel._prepend(value, this);
       _elementCount++;
     }
   }
 
   E removeLast() {
-    _DoubleLinkedQueueEntry<E> lastEntry = _sentinel._previousLink;
-    E result = lastEntry._remove();
+    // Hits sentinel's `_remove` if queue is empty.
+    E result = _sentinel._previousLink!._remove();
     _elementCount--;
     return result;
   }
 
   E removeFirst() {
-    _DoubleLinkedQueueEntry<E> firstEntry = _sentinel._nextLink;
-    E result = firstEntry._remove();
+    // Hits sentinel's `_remove` if queue is empty.
+    E result = _sentinel._nextLink!._remove();
     _elementCount--;
     return result;
   }
 
-  bool remove(Object o) {
-    _DoubleLinkedQueueEntry<E> entry = _sentinel._nextLink;
-    while (!identical(entry, _sentinel)) {
-      bool equals = (entry._element == o);
-      if (!identical(this, entry._queue)) {
+  bool remove(Object? o) {
+    _DoubleLinkedQueueEntry<E> entry = _sentinel._nextLink!;
+    while (true) {
+      var elementEntry = entry._asNonSentinelEntry();
+      if (elementEntry == null) return false;
+      bool equals = (elementEntry.element == o);
+      if (!identical(this, elementEntry._queue)) {
         // Entry must still be in the queue.
         throw ConcurrentModificationError(this);
       }
@@ -347,22 +353,23 @@ class DoubleLinkedQueue<E> extends Iterable<E> implements Queue<E> {
         _elementCount--;
         return true;
       }
-      entry = entry._nextLink;
+      entry = entry._nextLink!;
     }
-    return false;
   }
 
   void _filter(bool test(E element), bool removeMatching) {
-    _DoubleLinkedQueueEntry<E> entry = _sentinel._nextLink;
-    while (!identical(entry, _sentinel)) {
-      bool matches = test(entry._element);
-      if (!identical(this, entry._queue)) {
+    _DoubleLinkedQueueEntry<E> entry = _sentinel._nextLink!;
+    while (true) {
+      var elementEntry = entry._asNonSentinelEntry();
+      if (elementEntry == null) return;
+      bool matches = test(elementEntry.element);
+      if (!identical(this, elementEntry._queue)) {
         // Entry must still be in the queue.
         throw ConcurrentModificationError(this);
       }
-      _DoubleLinkedQueueEntry<E> next = entry._nextLink; // Cannot be null.
+      var next = entry._nextLink!; // Cannot be null while entry is in queue.
       if (identical(removeMatching, matches)) {
-        entry._remove();
+        elementEntry._remove();
         _elementCount--;
       }
       entry = next;
@@ -377,22 +384,17 @@ class DoubleLinkedQueue<E> extends Iterable<E> implements Queue<E> {
     _filter(test, false);
   }
 
-  E get first {
-    _DoubleLinkedQueueEntry<E> firstEntry = _sentinel._nextLink;
-    return firstEntry._element;
-  }
+  // Hits sentinel's `get element` if no element in queue.
+  E get first => _sentinel._nextLink!.element;
 
-  E get last {
-    _DoubleLinkedQueueEntry<E> lastEntry = _sentinel._previousLink;
-    return lastEntry._element;
-  }
+  // Hits sentinel's `get element` if no element in queue.
+  E get last => _sentinel._previousLink!.element;
 
   E get single {
     // Note that this throws correctly if the queue is empty
     // because reading the element of the sentinel throws.
     if (identical(_sentinel._nextLink, _sentinel._previousLink)) {
-      _DoubleLinkedQueueEntry<E> entry = _sentinel._nextLink;
-      return entry._element;
+      return _sentinel._nextLink!.element;
     }
     throw IterableElementError.tooMany();
   }
@@ -400,32 +402,42 @@ class DoubleLinkedQueue<E> extends Iterable<E> implements Queue<E> {
   /// The entry object of the first element in the queue.
   ///
   /// Each element of the queue has an associated [DoubleLinkedQueueEntry].
-  /// Returns the entry object corresponding to the first element of the queue.
+  ///
+  /// Returns the entry object corresponding to the first element of the queue,
+  /// or `null` if the queue is empty.
   ///
   /// The entry objects can also be accessed using [lastEntry],
-  /// and they can be iterated using [DoubleLinkedQueueEntry.nextEntry()] and
-  /// [DoubleLinkedQueueEntry.previousEntry()].
-  DoubleLinkedQueueEntry<E> firstEntry() {
-    return _sentinel.nextEntry();
-  }
+  /// and they can be iterated using [DoubleLinkedQueueEntry.nextEntry] and
+  /// [DoubleLinkedQueueEntry.previousEntry].
+  DoubleLinkedQueueEntry<E>? firstEntry() =>
+      _sentinel._nextLink!._asNonSentinelEntry();
 
   /// The entry object of the last element in the queue.
   ///
   /// Each element of the queue has an associated [DoubleLinkedQueueEntry].
-  /// Returns the entry object corresponding to the last element of the queue.
+  ///
+  /// Returns the entry object corresponding to the last element of the queue,
+  /// or `null` if the queue is empty.
   ///
   /// The entry objects can also be accessed using [firstEntry],
-  /// and they can be iterated using [DoubleLinkedQueueEntry.nextEntry()] and
-  /// [DoubleLinkedQueueEntry.previousEntry()].
-  DoubleLinkedQueueEntry<E> lastEntry() {
-    return _sentinel.previousEntry();
-  }
+  /// and they can be iterated using [DoubleLinkedQueueEntry.nextEntry] and
+  /// [DoubleLinkedQueueEntry.previousEntry].
+  DoubleLinkedQueueEntry<E>? lastEntry() =>
+      _sentinel._previousLink!._asNonSentinelEntry();
 
-  bool get isEmpty {
-    return (identical(_sentinel._nextLink, _sentinel));
-  }
+  bool get isEmpty => identical(_sentinel._nextLink, _sentinel);
 
   void clear() {
+    var cursor = _sentinel._nextLink!;
+    while (true) {
+      var entry = cursor._asNonSentinelEntry();
+      if (entry == null) break;
+      cursor = cursor._nextLink!;
+      entry
+        .._nextLink = null
+        .._previousLink = null
+        .._queue = null;
+    }
     _sentinel._nextLink = _sentinel;
     _sentinel._previousLink = _sentinel;
     _elementCount = 0;
@@ -442,63 +454,69 @@ class DoubleLinkedQueue<E> extends Iterable<E> implements Queue<E> {
   /// [DoubleLinkedQueueEntry.previousEntry()].
   ///
   /// The [action] function can use methods on [DoubleLinkedQueueEntry] to
-  /// remove the entry or it can insert elements before or after then entry.
+  /// remove the entry or it can insert elements before or after the entry.
   /// If the current entry is removed, iteration continues with the entry that
   /// was following the current entry when [action] was called. Any elements
   /// inserted after the current element before it is removed will not be
   /// visited by the iteration.
   void forEachEntry(void action(DoubleLinkedQueueEntry<E> element)) {
-    _DoubleLinkedQueueEntry<E> entry = _sentinel._nextLink;
-    while (!identical(entry, _sentinel)) {
-      _DoubleLinkedQueueElement<E> element = entry;
-      _DoubleLinkedQueueEntry<E> next = element._nextLink;
-      // Remember both entry and entry._nextLink.
-      // If someone calls `element.remove()` we continue from `next`.
-      // Otherwise we use the value of entry._nextLink which may have been
-      // updated.
-      action(element);
-      if (identical(this, entry._queue)) {
-        next = entry._nextLink;
-      } else if (!identical(this, next._queue)) {
+    var cursor = _sentinel._nextLink!;
+    while (true) {
+      var element = cursor._asNonSentinelEntry();
+      if (element == null) break;
+      if (!identical(element._queue, this)) {
         throw ConcurrentModificationError(this);
       }
-      entry = next;
+      cursor = cursor._nextLink!;
+      // Remember both element and element._nextLink (as cursor).
+      // If someone calls `element.remove()` we continue from `next`.
+      // Otherwise we use the value of element._nextLink which may have been
+      // updated.
+      action(element);
+      if (identical(this, element._queue)) {
+        cursor = element._nextLink!;
+      }
     }
   }
 
   _DoubleLinkedQueueIterator<E> get iterator {
-    return _DoubleLinkedQueueIterator<E>(_sentinel);
+    return _DoubleLinkedQueueIterator<E>(this);
   }
 
-  String toString() => IterableBase.iterableToFullString(this, '{', '}');
+  String toString() => Iterable.iterableToFullString(this, '{', '}');
 }
 
 class _DoubleLinkedQueueIterator<E> implements Iterator<E> {
-  _DoubleLinkedQueueSentinel<E> _sentinel;
-  _DoubleLinkedQueueEntry<E> _nextEntry;
-  E _current;
+  /// Queue being iterated. Used for concurrent modification checks.
+  DoubleLinkedQueue<E>? _queue;
 
-  _DoubleLinkedQueueIterator(_DoubleLinkedQueueSentinel<E> sentinel)
-      : _sentinel = sentinel,
-        _nextEntry = sentinel._nextLink;
+  /// Next entry to visit. Set to null when hitting the sentinel.
+  _DoubleLinkedQueueEntry<E>? _nextEntry;
+
+  /// Current element value, when valid.
+  E? _current;
+
+  _DoubleLinkedQueueIterator(DoubleLinkedQueue<E> this._queue)
+      : _nextEntry = _queue._sentinel._nextLink;
 
   bool moveNext() {
-    if (identical(_nextEntry, _sentinel)) {
+    var nextElement = _nextEntry?._asNonSentinelEntry();
+    if (nextElement == null) {
+      // Clear everything to not unnecessarily keep values alive.
       _current = null;
       _nextEntry = null;
-      _sentinel = null;
+      _queue = null;
       return false;
     }
-    _DoubleLinkedQueueElement<E> elementEntry = _nextEntry;
-    if (!identical(_sentinel._queue, elementEntry._queue)) {
-      throw ConcurrentModificationError(_sentinel._queue);
+    if (!identical(_queue, nextElement._queue)) {
+      throw ConcurrentModificationError(_queue);
     }
-    _current = elementEntry._element;
-    _nextEntry = elementEntry._nextLink;
+    _current = nextElement.element;
+    _nextEntry = nextElement._nextLink;
     return true;
   }
 
-  E get current => _current;
+  E get current => _current as E;
 }
 
 /// List based [Queue].
@@ -508,9 +526,66 @@ class _DoubleLinkedQueueIterator<E> implements Iterator<E> {
 /// amortized constant time add operations.
 ///
 /// The structure is efficient for any queue or stack usage.
-class ListQueue<E> extends ListIterable<E> implements Queue<E> {
+///
+/// Example:
+/// ```dart
+/// final queue = ListQueue<int>();
+/// ```
+/// To add objects to a queue, use [add], [addAll], [addFirst] or[addLast].
+/// ```dart continued
+/// queue.add(5);
+/// queue.addFirst(0);
+/// queue.addLast(10);
+/// queue.addAll([1, 2, 3]);
+/// print(queue); // {0, 5, 10, 1, 2, 3}
+/// ```
+/// To check if the queue is empty, use [isEmpty] or [isNotEmpty].
+/// To find the number of queue entries, use [length].
+/// ```dart continued
+/// final isEmpty = queue.isEmpty; // false
+/// final queueSize = queue.length; // 6
+/// ```
+/// To get first or last item from queue, use [first] or [last].
+/// ```dart continued
+/// final first = queue.first; // 0
+/// final last = queue.last; // 3
+/// ```
+/// To get item value using index, use [elementAt].
+/// ```dart continued
+/// final itemAt = queue.elementAt(2); // 10
+/// ```
+/// To convert queue to list, call [toList].
+/// ```dart continued
+/// final numbers = queue.toList();
+/// print(numbers); // [0, 5, 10, 1, 2, 3]
+/// ```
+/// To remove item from queue, call [remove], [removeFirst] or [removeLast].
+/// ```dart continued
+/// queue.remove(10);
+/// queue.removeFirst();
+/// queue.removeLast();
+/// print(queue); // {5, 1, 2}
+/// ```
+/// To remove multiple elements at the same time, use [removeWhere].
+/// ```dart continued
+/// queue.removeWhere((element) => element == 1);
+/// print(queue); // {5, 2}
+/// ```
+/// To remove all elements in this queue that do not meet a condition,
+/// use [retainWhere].
+/// ```dart continued
+/// queue.retainWhere((element) => element < 4);
+/// print(queue); // {2}
+/// ```
+/// To remove all items and empty the set, use [clear].
+/// ```dart continued
+/// queue.clear();
+/// print(queue.isEmpty); // true
+/// print(queue); // {}
+/// ```
+final class ListQueue<E> extends ListIterable<E> implements Queue<E> {
   static const int _INITIAL_CAPACITY = 8;
-  List<E> _table;
+  List<E?> _table;
   int _head;
   int _tail;
   int _modificationCount = 0;
@@ -519,16 +594,19 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
   ///
   /// If [initialCapacity] is given, prepare the queue for at least that many
   /// elements.
-  ListQueue([int initialCapacity])
+  ListQueue([int? initialCapacity])
       : _head = 0,
-        _tail = 0 {
+        _tail = 0,
+        _table = List<E?>.filled(_calculateCapacity(initialCapacity), null);
+
+  static int _calculateCapacity(int? initialCapacity) {
     if (initialCapacity == null || initialCapacity < _INITIAL_CAPACITY) {
-      initialCapacity = _INITIAL_CAPACITY;
+      return _INITIAL_CAPACITY;
     } else if (!_isPowerOf2(initialCapacity)) {
-      initialCapacity = _nextPowerOf2(initialCapacity);
+      return _nextPowerOf2(initialCapacity);
     }
     assert(_isPowerOf2(initialCapacity));
-    _table = List<E>(initialCapacity);
+    return initialCapacity;
   }
 
   /// Create a `ListQueue` containing all [elements].
@@ -542,15 +620,21 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
   /// ```dart
   /// Queue<SuperType> superQueue = ...;
   /// Queue<SubType> subQueue =
-  ///     new ListQueue<SubType>.from(superQueue.whereType<SubType>());
+  ///     ListQueue<SubType>.from(superQueue.whereType<SubType>());
   /// ```
-  factory ListQueue.from(Iterable elements) {
-    if (elements is List) {
+  /// Example:
+  /// ```dart
+  /// final numbers = <num>[10, 20, 30];
+  /// final queue = ListQueue<int>.from(numbers);
+  /// print(queue); // {10, 20, 30}
+  /// ```
+  factory ListQueue.from(Iterable<dynamic> elements) {
+    if (elements is List<dynamic>) {
       int length = elements.length;
       ListQueue<E> queue = ListQueue<E>(length + 1);
       assert(queue._table.length > length);
       for (int i = 0; i < length; i++) {
-        queue._table[i] = elements[i];
+        queue._table[i] = elements[i] as E;
       }
       queue._tail = length;
       return queue;
@@ -561,7 +645,7 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
       }
       ListQueue<E> result = ListQueue<E>(capacity);
       for (final element in elements) {
-        result.addLast(element);
+        result.addLast(element as E);
       }
       return result;
     }
@@ -571,6 +655,12 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
   ///
   /// The elements are added to the queue, as by [addLast], in the order given
   /// by `elements.iterator`.
+  /// Example:
+  /// ```dart
+  /// final baseQueue = ListQueue.of([1.0, 2.0, 3.0]); // A ListQueue<double>
+  /// final numQueue = ListQueue<num>.of(baseQueue);
+  /// print(numQueue); // {1.0, 2.0, 3.0}
+  /// ```
   factory ListQueue.of(Iterable<E> elements) =>
       ListQueue<E>()..addAll(elements);
 
@@ -582,7 +672,7 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
   void forEach(void f(E element)) {
     int modificationCount = _modificationCount;
     for (int i = _head; i != _tail; i = (i + 1) & (_table.length - 1)) {
-      f(_table[i]);
+      f(_table[i] as E);
       _checkModification(modificationCount);
     }
   }
@@ -593,33 +683,34 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
 
   E get first {
     if (_head == _tail) throw IterableElementError.noElement();
-    return _table[_head];
+    return _table[_head] as E;
   }
 
   E get last {
     if (_head == _tail) throw IterableElementError.noElement();
-    return _table[(_tail - 1) & (_table.length - 1)];
+    return _table[(_tail - 1) & (_table.length - 1)] as E;
   }
 
   E get single {
     if (_head == _tail) throw IterableElementError.noElement();
     if (length > 1) throw IterableElementError.tooMany();
-    return _table[_head];
+    return _table[_head] as E;
   }
 
   E elementAt(int index) {
-    RangeError.checkValidIndex(index, this);
-    return _table[(_head + index) & (_table.length - 1)];
+    IndexError.check(index, length, indexable: this);
+    return _table[(_head + index) & (_table.length - 1)] as E;
   }
 
   List<E> toList({bool growable = true}) {
-    List<E> list;
-    if (growable) {
-      list = <E>[]..length = length;
-    } else {
-      list = List<E>(length);
+    int mask = _table.length - 1;
+    int length = (_tail - _head) & mask;
+    if (length == 0) return List<E>.empty(growable: growable);
+
+    var list = List<E>.filled(length, first, growable: growable);
+    for (int i = 0; i < length; i++) {
+      list[i] = _table[(_head + i) & mask] as E;
     }
-    _writeToList(list);
     return list;
   }
 
@@ -658,9 +749,9 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
     }
   }
 
-  bool remove(Object value) {
+  bool remove(Object? value) {
     for (int i = _head; i != _tail; i = (i + 1) & (_table.length - 1)) {
-      E element = _table[i];
+      E? element = _table[i];
       if (element == value) {
         _remove(i);
         _modificationCount++;
@@ -674,7 +765,7 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
     int modificationCount = _modificationCount;
     int i = _head;
     while (i != _tail) {
-      E element = _table[i];
+      E element = _table[i] as E;
       bool remove = identical(removeMatching, test(element));
       _checkModification(modificationCount);
       if (remove) {
@@ -712,7 +803,7 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
     }
   }
 
-  String toString() => IterableBase.iterableToFullString(this, "{", "}");
+  String toString() => Iterable.iterableToFullString(this, "{", "}");
 
   // Queue interface.
 
@@ -730,7 +821,7 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
   E removeFirst() {
     if (_head == _tail) throw IterableElementError.noElement();
     _modificationCount++;
-    E result = _table[_head];
+    E result = _table[_head] as E;
     _table[_head] = null;
     _head = (_head + 1) & (_table.length - 1);
     return result;
@@ -740,7 +831,7 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
     if (_head == _tail) throw IterableElementError.noElement();
     _modificationCount++;
     _tail = (_tail - 1) & (_table.length - 1);
-    E result = _table[_tail];
+    E result = _table[_tail] as E;
     _table[_tail] = null;
     return result;
   }
@@ -820,7 +911,7 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
 
   /// Grow the table when full.
   void _grow() {
-    List<E> newTable = List<E>(_table.length * 2);
+    List<E?> newTable = List<E?>.filled(_table.length * 2, null);
     int split = _table.length - _head;
     newTable.setRange(0, split, _table, _head);
     newTable.setRange(split, split + _head, _table, 0);
@@ -829,7 +920,7 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
     _table = newTable;
   }
 
-  int _writeToList(List<E> target) {
+  int _writeToList(List<E?> target) {
     assert(target.length >= length);
     if (_head <= _tail) {
       int length = _tail - _head;
@@ -851,7 +942,7 @@ class ListQueue<E> extends ListIterable<E> implements Queue<E> {
     // expansion.
     newElementCount += newElementCount >> 1;
     int newCapacity = _nextPowerOf2(newElementCount);
-    List<E> newTable = List<E>(newCapacity);
+    List<E?> newTable = List<E?>.filled(newCapacity, null);
     _tail = _writeToList(newTable);
     _table = newTable;
     _head = 0;
@@ -866,7 +957,7 @@ class _ListQueueIterator<E> implements Iterator<E> {
   final int _end;
   final int _modificationCount;
   int _position;
-  E _current;
+  E? _current;
 
   _ListQueueIterator(ListQueue<E> queue)
       : _queue = queue,
@@ -874,7 +965,7 @@ class _ListQueueIterator<E> implements Iterator<E> {
         _modificationCount = queue._modificationCount,
         _position = queue._head;
 
-  E get current => _current;
+  E get current => _current as E;
 
   bool moveNext() {
     _queue._checkModification(_modificationCount);
