@@ -2,8 +2,6 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file.
 
-// @dart = 2.6
-
 part of dart.convert;
 
 /// Error thrown by JSON serialization if an object cannot be serialized.
@@ -16,15 +14,15 @@ part of dart.convert;
 /// serializable, the [cause] is null.
 class JsonUnsupportedObjectError extends Error {
   /// The object that could not be serialized.
-  final Object unsupportedObject;
+  final Object? unsupportedObject;
 
   /// The exception thrown when trying to convert the object.
-  final Object cause;
+  final Object? cause;
 
   /// The partial result of the conversion, up until the error happened.
   ///
   /// May be null.
-  final String partialResult;
+  final String? partialResult;
 
   JsonUnsupportedObjectError(this.unsupportedObject,
       {this.cause, this.partialResult});
@@ -48,7 +46,7 @@ class JsonUnsupportedObjectError extends Error {
 /// When the cycle is detected, a [JsonCyclicError] is thrown.
 class JsonCyclicError extends JsonUnsupportedObjectError {
   /// The first object that was detected as part of a cycle.
-  JsonCyclicError(Object object) : super(object);
+  JsonCyclicError(Object? object) : super(object);
   String toString() => "Cyclic error in JSON stringify";
 }
 
@@ -58,15 +56,15 @@ class JsonCyclicError extends JsonUnsupportedObjectError {
 /// use cases.
 ///
 /// Examples:
-///
-///     var encoded = json.encode([1, 2, { "a": null }]);
-///     var decoded = json.decode('["foo", { "bar": 499 }]');
-///
+/// ```dart
+/// var encoded = json.encode([1, 2, { "a": null }]);
+/// var decoded = json.decode('["foo", { "bar": 499 }]');
+/// ```
 /// The top-level [jsonEncode] and [jsonDecode] functions may be used instead if
 /// a local variable shadows the [json] constant.
 const JsonCodec json = JsonCodec();
 
-/// Converts [value] to a JSON string.
+/// Converts [object] to a JSON string.
 ///
 /// If value contains objects that are not directly encodable to a JSON
 /// string (a value that is not a number, boolean, string, null, list or a map
@@ -78,8 +76,42 @@ const JsonCodec json = JsonCodec();
 ///
 /// Shorthand for `json.encode`. Useful if a local variable shadows the global
 /// [json] constant.
-String jsonEncode(Object value, {Object toEncodable(Object nonEncodable)}) =>
-    json.encode(value, toEncodable: toEncodable);
+///
+/// Example:
+/// ```dart
+/// const data = {'text': 'foo', 'value': 2, 'status': false, 'extra': null};
+/// final String jsonString = jsonEncode(data);
+/// print(jsonString); // {"text":"foo","value":2,"status":false,"extra":null}
+/// ```
+///
+/// Example of converting an otherwise unsupported object to a
+/// custom JSON format:
+///
+/// ```dart
+/// class CustomClass {
+///   final String text;
+///   final int value;
+///   CustomClass({required this.text, required this.value});
+///   CustomClass.fromJson(Map<String, dynamic> json)
+///       : text = json['text'],
+///         value = json['value'];
+///
+///   static Map<String, dynamic> toJson(CustomClass value) =>
+///       {'text': value.text, 'value': value.value};
+/// }
+///
+/// void main() {
+///   final CustomClass cc = CustomClass(text: 'Dart', value: 123);
+///   final jsonText = jsonEncode({'cc': cc},
+///       toEncodable: (Object? value) => value is CustomClass
+///           ? CustomClass.toJson(value)
+///           : throw UnsupportedError('Cannot convert to JSON: $value'));
+///   print(jsonText); // {"cc":{"text":"Dart","value":123}}
+/// }
+/// ```
+String jsonEncode(Object? object,
+        {Object? toEncodable(Object? nonEncodable)?}) =>
+    json.encode(object, toEncodable: toEncodable);
 
 /// Parses the string and returns the resulting Json object.
 ///
@@ -92,19 +124,47 @@ String jsonEncode(Object value, {Object toEncodable(Object nonEncodable)}) =>
 ///
 /// Shorthand for `json.decode`. Useful if a local variable shadows the global
 /// [json] constant.
-dynamic jsonDecode(String source, {Object reviver(Object key, Object value)}) =>
+///
+/// Example:
+/// ```dart
+/// const jsonString =
+///     '{"text": "foo", "value": 1, "status": false, "extra": null}';
+///
+/// final data = jsonDecode(jsonString);
+/// print(data['text']); // foo
+/// print(data['value']); // 1
+/// print(data['status']); // false
+/// print(data['extra']); // null
+///
+/// const jsonArray = '''
+///   [{"text": "foo", "value": 1, "status": true},
+///    {"text": "bar", "value": 2, "status": false}]
+/// ''';
+///
+/// final List<dynamic> dataList = jsonDecode(jsonArray);
+/// print(dataList[0]); // {text: foo, value: 1, status: true}
+/// print(dataList[1]); // {text: bar, value: 2, status: false}
+///
+/// final item = dataList[0];
+/// print(item['text']); // foo
+/// print(item['value']); // 1
+/// print(item['status']); // false
+/// ```
+dynamic jsonDecode(String source,
+        {Object? reviver(Object? key, Object? value)?}) =>
     json.decode(source, reviver: reviver);
 
 /// A [JsonCodec] encodes JSON objects to strings and decodes strings to
 /// JSON objects.
 ///
 /// Examples:
-///
-///     var encoded = json.encode([1, 2, { "a": null }]);
-///     var decoded = json.decode('["foo", { "bar": 499 }]');
-class JsonCodec extends Codec<Object, String> {
-  final Function(Object key, Object value) _reviver;
-  final Function(dynamic) _toEncodable;
+/// ```dart
+/// var encoded = json.encode([1, 2, { "a": null }]);
+/// var decoded = json.decode('["foo", { "bar": 499 }]');
+/// ```
+final class JsonCodec extends Codec<Object?, String> {
+  final Object? Function(Object? key, Object? value)? _reviver;
+  final Object? Function(dynamic)? _toEncodable;
 
   /// Creates a `JsonCodec` with the given reviver and encoding function.
   ///
@@ -127,7 +187,9 @@ class JsonCodec extends Codec<Object, String> {
   ///
   /// If [toEncodable] is omitted, it defaults to a function that returns the
   /// result of calling `.toJson()` on the unencodable object.
-  const JsonCodec({reviver(Object key, Object value), toEncodable(var object)})
+  const JsonCodec(
+      {Object? reviver(Object? key, Object? value)?,
+      Object? toEncodable(dynamic object)?})
       : _reviver = reviver,
         _toEncodable = toEncodable;
 
@@ -137,7 +199,7 @@ class JsonCodec extends Codec<Object, String> {
   /// that has been parsed during decoding. The `key` argument is either the
   /// integer list index for a list property, the string map key for object
   /// properties, or `null` for the final result.
-  JsonCodec.withReviver(reviver(Object key, Object value))
+  JsonCodec.withReviver(dynamic reviver(Object? key, Object? value))
       : this(reviver: reviver);
 
   /// Parses the string and returns the resulting Json object.
@@ -148,7 +210,8 @@ class JsonCodec extends Codec<Object, String> {
   /// properties, or `null` for the final result.
   ///
   /// The default [reviver] (when not provided) is the identity function.
-  dynamic decode(String source, {reviver(Object key, Object value)}) {
+  dynamic decode(String source,
+      {Object? reviver(Object? key, Object? value)?}) {
     reviver ??= _reviver;
     if (reviver == null) return decoder.convert(source);
     return JsonDecoder(reviver).convert(source);
@@ -163,7 +226,7 @@ class JsonCodec extends Codec<Object, String> {
   ///
   /// If [toEncodable] is omitted, it defaults to a function that returns the
   /// result of calling `.toJson()` on the unencodable object.
-  String encode(Object value, {toEncodable(object)}) {
+  String encode(Object? value, {Object? toEncodable(dynamic object)?}) {
     toEncodable ??= _toEncodable;
     if (toEncodable == null) return encoder.convert(value);
     return JsonEncoder(toEncodable).convert(value);
@@ -181,18 +244,42 @@ class JsonCodec extends Codec<Object, String> {
 }
 
 /// This class converts JSON objects to strings.
-class JsonEncoder extends Converter<Object, String> {
+///
+/// Example:
+///
+/// ```dart
+/// const JsonEncoder encoder = JsonEncoder();
+/// const data = {'text': 'foo', 'value': '2'};
+///
+/// final String jsonString = encoder.convert(data);
+/// print(jsonString); // {"text":"foo","value":"2"}
+/// ```
+///
+/// Example of pretty-printed output:
+///
+/// ```dart
+/// const JsonEncoder encoder = JsonEncoder.withIndent('  ');
+///
+/// const data = {'text': 'foo', 'value': '2'};
+/// final String jsonString = encoder.convert(data);
+/// print(jsonString);
+/// // {
+/// //   "text": "foo",
+/// //   "value": "2"
+/// // }
+/// ```
+final class JsonEncoder extends Converter<Object?, String> {
   /// The string used for indention.
   ///
   /// When generating multi-line output, this string is inserted once at the
   /// beginning of each indented line for each level of indentation.
   ///
   /// If `null`, the output is encoded as a single line.
-  final String indent;
+  final String? indent;
 
   /// Function called on non-encodable objects to return a replacement
   /// encodable object that will be encoded in the orignal's place.
-  final Function(dynamic) _toEncodable;
+  final Object? Function(dynamic)? _toEncodable;
 
   /// Creates a JSON encoder.
   ///
@@ -204,7 +291,7 @@ class JsonEncoder extends Converter<Object, String> {
   ///
   /// If [toEncodable] is omitted, it defaults to calling `.toJson()` on
   /// the object.
-  const JsonEncoder([toEncodable(object)])
+  const JsonEncoder([Object? toEncodable(dynamic object)?])
       : indent = null,
         _toEncodable = toEncodable;
 
@@ -224,7 +311,8 @@ class JsonEncoder extends Converter<Object, String> {
   ///
   /// If [toEncodable] is omitted, it defaults to calling `.toJson()` on
   /// the object.
-  const JsonEncoder.withIndent(this.indent, [toEncodable(object)])
+  const JsonEncoder.withIndent(this.indent,
+      [Object? toEncodable(dynamic object)?])
       : _toEncodable = toEncodable;
 
   /// Converts [object] to a JSON [String].
@@ -242,7 +330,7 @@ class JsonEncoder extends Converter<Object, String> {
   /// If the conversion throws, or returns a value that is not directly
   /// serializable, a [JsonUnsupportedObjectError] exception is thrown.
   /// If the call throws, the error is caught and stored in the
-  /// [JsonUnsupportedObjectError]'s [:cause:] field.
+  /// [JsonUnsupportedObjectError]'s `cause` field.
   ///
   /// If a [List] or [Map] contains a reference to itself, directly or through
   /// other lists or maps, it cannot be serialized and a [JsonCyclicError] is
@@ -253,7 +341,7 @@ class JsonEncoder extends Converter<Object, String> {
   /// If an object is serialized more than once, [convert] may cache the text
   /// for it. In other words, if the content of an object changes after it is
   /// first serialized, the new values may not be reflected in the result.
-  String convert(Object object) =>
+  String convert(Object? object) =>
       _JsonStringStringifier.stringify(object, _toEncodable, indent);
 
   /// Starts a chunked conversion.
@@ -263,7 +351,7 @@ class JsonEncoder extends Converter<Object, String> {
   ///
   /// Returns a chunked-conversion sink that accepts at most one object. It is
   /// an error to invoke `add` more than once on the returned sink.
-  ChunkedConversionSink<Object> startChunkedConversion(Sink<String> sink) {
+  ChunkedConversionSink<Object?> startChunkedConversion(Sink<String> sink) {
     if (sink is _Utf8EncoderSink) {
       return _JsonUtf8EncoderSink(
           sink._sink,
@@ -278,15 +366,13 @@ class JsonEncoder extends Converter<Object, String> {
   }
 
   // Override the base class's bind, to provide a better type.
-  Stream<String> bind(Stream<Object> stream) => super.bind(stream);
+  Stream<String> bind(Stream<Object?> stream) => super.bind(stream);
 
-  Converter<Object, T> fuse<T>(Converter<String, T> other) {
+  Converter<Object?, T> fuse<T>(Converter<String, T> other) {
     if (other is Utf8Encoder) {
       // The instance check guarantees that `T` is (a subtype of) List<int>,
       // but the static type system doesn't know that, and so we cast.
-      // Cast through dynamic to keep the cast implicit for builds using
-      // unchecked implicit casts.
-      return JsonUtf8Encoder(indent, _toEncodable) as dynamic;
+      return JsonUtf8Encoder(indent, _toEncodable) as Converter<Object?, T>;
     }
     return super.fuse<T>(other);
   }
@@ -297,17 +383,15 @@ class JsonEncoder extends Converter<Object, String> {
 /// This encoder works equivalently to first converting the object to
 /// a JSON string, and then UTF-8 encoding the string, but without
 /// creating an intermediate string.
-class JsonUtf8Encoder extends Converter<Object, List<int>> {
+final class JsonUtf8Encoder extends Converter<Object?, List<int>> {
   /// Default buffer size used by the JSON-to-UTF-8 encoder.
   static const int _defaultBufferSize = 256;
-  @deprecated
-  static const int DEFAULT_BUFFER_SIZE = _defaultBufferSize;
 
   /// Indentation used in pretty-print mode, `null` if not pretty.
-  final List<int> _indent;
+  final List<int>? _indent;
 
   /// Function called with each un-encodable object encountered.
-  final Function(dynamic) _toEncodable;
+  final Object? Function(dynamic)? _toEncodable;
 
   /// UTF-8 buffer size.
   final int _bufferSize;
@@ -336,12 +420,12 @@ class JsonUtf8Encoder extends Converter<Object, List<int>> {
   /// If [toEncodable] is omitted, it defaults to calling `.toJson()` on the
   /// object.
   JsonUtf8Encoder(
-      [String indent, toEncodable(object), int bufferSize = _defaultBufferSize])
+      [String? indent, dynamic toEncodable(dynamic object)?, int? bufferSize])
       : _indent = _utf8Encode(indent),
         _toEncodable = toEncodable,
-        _bufferSize = bufferSize;
+        _bufferSize = bufferSize ?? _defaultBufferSize;
 
-  static List<int> _utf8Encode(String string) {
+  static List<int>? _utf8Encode(String? string) {
     if (string == null) return null;
     if (string.isEmpty) return Uint8List(0);
     checkAscii:
@@ -355,7 +439,7 @@ class JsonUtf8Encoder extends Converter<Object, List<int>> {
   }
 
   /// Convert [object] into UTF-8 encoded JSON.
-  List<int> convert(Object object) {
+  List<int> convert(Object? object) {
     var bytes = <List<int>>[];
     // The `stringify` function always converts into chunks.
     // Collect the chunks into the `bytes` list, then combine them afterwards.
@@ -391,7 +475,7 @@ class JsonUtf8Encoder extends Converter<Object, List<int>> {
   ///
   /// The argument [sink] will receive byte lists in sizes depending on the
   /// `bufferSize` passed to the constructor when creating this encoder.
-  ChunkedConversionSink<Object> startChunkedConversion(Sink<List<int>> sink) {
+  ChunkedConversionSink<Object?> startChunkedConversion(Sink<List<int>> sink) {
     ByteConversionSink byteSink;
     if (sink is ByteConversionSink) {
       byteSink = sink;
@@ -402,7 +486,7 @@ class JsonUtf8Encoder extends Converter<Object, List<int>> {
   }
 
   // Override the base class's bind, to provide a better type.
-  Stream<List<int>> bind(Stream<Object> stream) {
+  Stream<List<int>> bind(Stream<Object?> stream) {
     return super.bind(stream);
   }
 }
@@ -410,9 +494,9 @@ class JsonUtf8Encoder extends Converter<Object, List<int>> {
 /// Implements the chunked conversion from object to its JSON representation.
 ///
 /// The sink only accepts one value, but will produce output in a chunked way.
-class _JsonEncoderSink extends ChunkedConversionSink<Object> {
-  final String _indent;
-  final Function(dynamic) _toEncodable;
+class _JsonEncoderSink extends ChunkedConversionSink<Object?> {
+  final String? _indent;
+  final Object? Function(dynamic)? _toEncodable;
   final StringConversionSink _sink;
   bool _isDone = false;
 
@@ -423,7 +507,7 @@ class _JsonEncoderSink extends ChunkedConversionSink<Object> {
   /// It is an error to invoke this method more than once on any instance. While
   /// this makes the input effectively non-chunked the output will be generated
   /// in a chunked way.
-  void add(Object o) {
+  void add(Object? o) {
     if (_isDone) {
       throw StateError("Only one call to add allowed");
     }
@@ -437,11 +521,11 @@ class _JsonEncoderSink extends ChunkedConversionSink<Object> {
 }
 
 /// Sink returned when starting a chunked conversion from object to bytes.
-class _JsonUtf8EncoderSink extends ChunkedConversionSink<Object> {
-  /// The byte sink receiveing the encoded chunks.
+class _JsonUtf8EncoderSink extends ChunkedConversionSink<Object?> {
+  /// The byte sink receiving the encoded chunks.
   final ByteConversionSink _sink;
-  final List<int> _indent;
-  final Function(dynamic) _toEncodable;
+  final List<int>? _indent;
+  final Object? Function(dynamic)? _toEncodable;
   final int _bufferSize;
   bool _isDone = false;
   _JsonUtf8EncoderSink(
@@ -452,7 +536,7 @@ class _JsonUtf8EncoderSink extends ChunkedConversionSink<Object> {
     _sink.addSlice(chunk, start, end, false);
   }
 
-  void add(Object object) {
+  void add(Object? object) {
     if (_isDone) {
       throw StateError("Only one call to add allowed");
     }
@@ -471,13 +555,44 @@ class _JsonUtf8EncoderSink extends ChunkedConversionSink<Object> {
 }
 
 /// This class parses JSON strings and builds the corresponding objects.
-class JsonDecoder extends Converter<String, Object> {
-  final Function(Object key, Object value) _reviver;
+///
+/// A JSON input must be the JSON encoding of a single JSON value,
+/// which can be a list or map containing other values.
+///
+/// Throws [FormatException] if the input is not valid JSON text.
+///
+/// Example:
+/// ```dart
+/// const JsonDecoder decoder = JsonDecoder();
+///
+/// const String jsonString = '''
+///   {
+///     "data": [{"text": "foo", "value": 1 },
+///              {"text": "bar", "value": 2 }],
+///     "text": "Dart"
+///   }
+/// ''';
+///
+/// final Map<String, dynamic> object = decoder.convert(jsonString);
+///
+/// final item = object['data'][0];
+/// print(item['text']); // foo
+/// print(item['value']); // 1
+///
+/// print(object['text']); // Dart
+/// ```
+///
+/// When used as a [StreamTransformer], the input stream may emit
+/// multiple strings. The concatenation of all of these strings must
+/// be a valid JSON encoding of a single JSON value.
+final class JsonDecoder extends Converter<String, Object?> {
+  final Object? Function(Object? key, Object? value)? _reviver;
 
   /// Constructs a new JsonDecoder.
   ///
   /// The [reviver] may be `null`.
-  const JsonDecoder([reviver(Object key, Object value)]) : _reviver = reviver;
+  const JsonDecoder([Object? reviver(Object? key, Object? value)?])
+      : _reviver = reviver;
 
   /// Converts the given JSON-string [input] to its corresponding object.
   ///
@@ -497,17 +612,18 @@ class JsonDecoder extends Converter<String, Object> {
   /// Starts a conversion from a chunked JSON string to its corresponding object.
   ///
   /// The output [sink] receives exactly one decoded element through `add`.
-  external StringConversionSink startChunkedConversion(Sink<Object> sink);
+  external StringConversionSink startChunkedConversion(Sink<Object?> sink);
 
   // Override the base class's bind, to provide a better type.
-  Stream<Object> bind(Stream<String> stream) => super.bind(stream);
+  Stream<Object?> bind(Stream<String> stream) => super.bind(stream);
 }
 
 // Internal optimized JSON parsing implementation.
-external _parseJson(String source, reviver(key, value));
+external dynamic _parseJson(String source, reviver(key, value)?);
 
 // Implementation of encoder/stringifier.
 
+// ignore: avoid_dynamic_calls
 dynamic _defaultToEncodable(dynamic object) => object.toJson();
 
 /// JSON encoder that traverses an object structure and writes JSON source.
@@ -525,11 +641,16 @@ abstract class _JsonStringifier {
   static const int char_0 = 0x30;
   static const int backslash = 0x5c;
   static const int char_b = 0x62;
+  static const int char_d = 0x64;
   static const int char_f = 0x66;
   static const int char_n = 0x6e;
   static const int char_r = 0x72;
   static const int char_t = 0x74;
   static const int char_u = 0x75;
+  static const int surrogateMin = 0xd800;
+  static const int surrogateMask = 0xfc00;
+  static const int surrogateLead = 0xd800;
+  static const int surrogateTrail = 0xdc00;
 
   /// List of objects currently being traversed. Used to detect cycles.
   final List _seen = [];
@@ -537,10 +658,10 @@ abstract class _JsonStringifier {
   /// Function called for each un-encodable object encountered.
   final Function(dynamic) _toEncodable;
 
-  _JsonStringifier(toEncodable(o))
+  _JsonStringifier(dynamic toEncodable(dynamic o)?)
       : _toEncodable = toEncodable ?? _defaultToEncodable;
 
-  String get _partialResult;
+  String? get _partialResult;
 
   /// Append a string to the JSON output.
   void writeString(String characters);
@@ -563,7 +684,30 @@ abstract class _JsonStringifier {
     final length = s.length;
     for (var i = 0; i < length; i++) {
       var charCode = s.codeUnitAt(i);
-      if (charCode > backslash) continue;
+      if (charCode > backslash) {
+        if (charCode >= surrogateMin) {
+          // Possible surrogate. Check if it is unpaired.
+          if (((charCode & surrogateMask) == surrogateLead &&
+                  !(i + 1 < length &&
+                      (s.codeUnitAt(i + 1) & surrogateMask) ==
+                          surrogateTrail)) ||
+              ((charCode & surrogateMask) == surrogateTrail &&
+                  !(i - 1 >= 0 &&
+                      (s.codeUnitAt(i - 1) & surrogateMask) ==
+                          surrogateLead))) {
+            // Lone surrogate.
+            if (i > offset) writeStringSlice(s, offset, i);
+            offset = i + 1;
+            writeCharCode(backslash);
+            writeCharCode(char_u);
+            writeCharCode(char_d);
+            writeCharCode(hexDigit((charCode >> 8) & 0xf));
+            writeCharCode(hexDigit((charCode >> 4) & 0xf));
+            writeCharCode(hexDigit(charCode & 0xf));
+          }
+        }
+        continue;
+      }
       if (charCode < 32) {
         if (i > offset) writeStringSlice(s, offset, i);
         offset = i + 1;
@@ -610,7 +754,7 @@ abstract class _JsonStringifier {
   ///
   /// Records the object if it isn't already seen. Should have a matching call to
   /// [_removeSeen] when the object is no longer being traversed.
-  void _checkCycle(object) {
+  void _checkCycle(Object? object) {
     for (var i = 0; i < _seen.length; i++) {
       if (identical(object, _seen[i])) {
         throw JsonCyclicError(object);
@@ -623,7 +767,7 @@ abstract class _JsonStringifier {
   ///
   /// Should be called in the opposite order of the matching [_checkCycle]
   /// calls.
-  void _removeSeen(object) {
+  void _removeSeen(Object? object) {
     assert(_seen.isNotEmpty);
     assert(identical(_seen.last, object));
     _seen.removeLast();
@@ -633,7 +777,7 @@ abstract class _JsonStringifier {
   ///
   /// If [object] isn't directly encodable, the [_toEncodable] function gets one
   /// chance to return a replacement which is encodable.
-  void writeObject(object) {
+  void writeObject(Object? object) {
     // Tries stringifying object directly. If it's not a simple value, List or
     // Map, call toJson() to get a custom representation and try serializing
     // that.
@@ -655,7 +799,7 @@ abstract class _JsonStringifier {
   ///
   /// Returns true if the value is one of these types, and false if not.
   /// If a value is both a [List] and a [Map], it's serialized as a [List].
-  bool writeJsonValue(object) {
+  bool writeJsonValue(Object? object) {
     if (object is num) {
       if (!object.isFinite) return false;
       writeNumber(object);
@@ -691,7 +835,7 @@ abstract class _JsonStringifier {
   }
 
   /// Serialize a [List].
-  void writeList(List list) {
+  void writeList(List<Object?> list) {
     writeString('[');
     if (list.isNotEmpty) {
       writeObject(list[0]);
@@ -704,12 +848,12 @@ abstract class _JsonStringifier {
   }
 
   /// Serialize a [Map].
-  bool writeMap(Map map) {
+  bool writeMap(Map<Object?, Object?> map) {
     if (map.isEmpty) {
       writeString("{}");
       return true;
     }
-    var keyValueList = List(map.length * 2);
+    var keyValueList = List<Object?>.filled(map.length * 2, null);
     var i = 0;
     var allStringKeys = true;
     map.forEach((key, value) {
@@ -725,7 +869,7 @@ abstract class _JsonStringifier {
     for (var i = 0; i < keyValueList.length; i += 2) {
       writeString(separator);
       separator = ',"';
-      writeStringContent(keyValueList[i]);
+      writeStringContent(keyValueList[i] as String);
       writeString('":');
       writeObject(keyValueList[i + 1]);
     }
@@ -738,13 +882,13 @@ abstract class _JsonStringifier {
 /// [Map] objects using the specified indent value.
 ///
 /// Subclasses should implement [writeIndentation].
-abstract class _JsonPrettyPrintMixin implements _JsonStringifier {
+mixin _JsonPrettyPrintMixin implements _JsonStringifier {
   int _indentLevel = 0;
 
   /// Add [indentLevel] indentations to the JSON output.
   void writeIndentation(int indentLevel);
 
-  void writeList(List list) {
+  void writeList(List<Object?> list) {
     if (list.isEmpty) {
       writeString('[]');
     } else {
@@ -764,12 +908,12 @@ abstract class _JsonPrettyPrintMixin implements _JsonStringifier {
     }
   }
 
-  bool writeMap(Map map) {
+  bool writeMap(Map<Object?, Object?> map) {
     if (map.isEmpty) {
       writeString("{}");
       return true;
     }
-    var keyValueList = List(map.length * 2);
+    var keyValueList = List<Object?>.filled(map.length * 2, null);
     var i = 0;
     var allStringKeys = true;
     map.forEach((key, value) {
@@ -788,7 +932,7 @@ abstract class _JsonPrettyPrintMixin implements _JsonStringifier {
       separator = ",\n";
       writeIndentation(_indentLevel);
       writeString('"');
-      writeStringContent(keyValueList[i]);
+      writeStringContent(keyValueList[i] as String);
       writeString('": ');
       writeObject(keyValueList[i + 1]);
     }
@@ -804,7 +948,8 @@ abstract class _JsonPrettyPrintMixin implements _JsonStringifier {
 class _JsonStringStringifier extends _JsonStringifier {
   final StringSink _sink;
 
-  _JsonStringStringifier(this._sink, dynamic Function(dynamic) _toEncodable)
+  _JsonStringStringifier(
+      this._sink, dynamic Function(dynamic object)? _toEncodable)
       : super(_toEncodable);
 
   /// Convert object to a string.
@@ -816,7 +961,8 @@ class _JsonStringStringifier extends _JsonStringifier {
   /// with newlines and indentation. The `indent` string is added as indentation
   /// for each indentation level. It should only contain valid JSON whitespace
   /// characters (space, tab, carriage return or line feed).
-  static String stringify(object, toEncodable(o), String indent) {
+  static String stringify(
+      Object? object, dynamic toEncodable(dynamic object)?, String? indent) {
     var output = StringBuffer();
     printOn(object, output, toEncodable, indent);
     return output.toString();
@@ -825,8 +971,8 @@ class _JsonStringStringifier extends _JsonStringifier {
   /// Convert object to a string, and write the result to the [output] sink.
   ///
   /// The result is written piecemally to the sink.
-  static void printOn(
-      object, StringSink output, toEncodable(o), String indent) {
+  static void printOn(Object? object, StringSink output,
+      dynamic toEncodable(dynamic o)?, String? indent) {
     _JsonStringifier stringifier;
     if (indent == null) {
       stringifier = _JsonStringStringifier(output, toEncodable);
@@ -836,7 +982,7 @@ class _JsonStringStringifier extends _JsonStringifier {
     stringifier.writeObject(object);
   }
 
-  String get _partialResult => _sink is StringBuffer ? _sink.toString() : null;
+  String? get _partialResult => _sink is StringBuffer ? _sink.toString() : null;
 
   void writeNumber(num number) {
     _sink.write(number.toString());
@@ -859,7 +1005,8 @@ class _JsonStringStringifierPretty extends _JsonStringStringifier
     with _JsonPrettyPrintMixin {
   final String _indent;
 
-  _JsonStringStringifierPretty(StringSink sink, toEncodable(o), this._indent)
+  _JsonStringStringifierPretty(
+      StringSink sink, dynamic toEncodable(dynamic o)?, this._indent)
       : super(sink, toEncodable);
 
   void writeIndentation(int count) {
@@ -877,7 +1024,8 @@ class _JsonUtf8Stringifier extends _JsonStringifier {
   Uint8List buffer;
   int index = 0;
 
-  _JsonUtf8Stringifier(toEncodable(o), this.bufferSize, this.addChunk)
+  _JsonUtf8Stringifier(
+      dynamic toEncodable(dynamic o)?, this.bufferSize, this.addChunk)
       : buffer = Uint8List(bufferSize),
         super(toEncodable);
 
@@ -890,8 +1038,12 @@ class _JsonUtf8Stringifier extends _JsonStringifier {
   ///
   /// If [indent] is non-`null`, the result will be "pretty-printed" with extra
   /// newlines and indentation, using [indent] as the indentation.
-  static void stringify(Object object, List<int> indent, toEncodable(o),
-      int bufferSize, void addChunk(Uint8List chunk, int start, int end)) {
+  static void stringify(
+      Object? object,
+      List<int>? indent,
+      dynamic toEncodable(dynamic o)?,
+      int bufferSize,
+      void addChunk(Uint8List chunk, int start, int end)) {
     _JsonUtf8Stringifier stringifier;
     if (indent != null) {
       stringifier =
@@ -909,11 +1061,11 @@ class _JsonUtf8Stringifier extends _JsonStringifier {
     if (index > 0) {
       addChunk(buffer, 0, index);
     }
-    buffer = null;
+    buffer = Uint8List(0);
     index = 0;
   }
 
-  String get _partialResult => null;
+  String? get _partialResult => null;
 
   void writeNumber(num number) {
     writeAsciiString(number.toString());
@@ -937,22 +1089,28 @@ class _JsonUtf8Stringifier extends _JsonStringifier {
   void writeStringSlice(String string, int start, int end) {
     // TODO(lrn): Optimize by copying directly into buffer instead of going
     // through writeCharCode/writeByte. Assumption is the most characters
-    // in starings are plain ASCII.
+    // in strings are plain ASCII.
     for (var i = start; i < end; i++) {
       var char = string.codeUnitAt(i);
       if (char <= 0x7f) {
         writeByte(char);
       } else {
-        if ((char & 0xFC00) == 0xD800 && i + 1 < end) {
-          // Lead surrogate.
-          var nextChar = string.codeUnitAt(i + 1);
-          if ((nextChar & 0xFC00) == 0xDC00) {
-            // Tail surrogate.
-            char = 0x10000 + ((char & 0x3ff) << 10) + (nextChar & 0x3ff);
-            writeFourByteCharCode(char);
-            i++;
-            continue;
+        if ((char & 0xF800) == 0xD800) {
+          // Surrogate.
+          if (char < 0xDC00 && i + 1 < end) {
+            // Lead surrogate.
+            var nextChar = string.codeUnitAt(i + 1);
+            if ((nextChar & 0xFC00) == 0xDC00) {
+              // Tail surrogate.
+              char = 0x10000 + ((char & 0x3ff) << 10) + (nextChar & 0x3ff);
+              writeFourByteCharCode(char);
+              i++;
+              continue;
+            }
           }
+          // Unpaired surrogate.
+          writeMultiByteCharCode(unicodeReplacementCharacterRune);
+          continue;
         }
         writeMultiByteCharCode(char);
       }
@@ -1005,8 +1163,8 @@ class _JsonUtf8Stringifier extends _JsonStringifier {
 class _JsonUtf8StringifierPretty extends _JsonUtf8Stringifier
     with _JsonPrettyPrintMixin {
   final List<int> indent;
-  _JsonUtf8StringifierPretty(toEncodable(o), this.indent, int bufferSize,
-      void addChunk(Uint8List buffer, int start, int end))
+  _JsonUtf8StringifierPretty(dynamic toEncodable(dynamic o)?, this.indent,
+      int bufferSize, void addChunk(Uint8List buffer, int start, int end))
       : super(toEncodable, bufferSize, addChunk);
 
   void writeIndentation(int count) {
